from delune.lib import unistr
import json

class Document:
	def __init__ (self, _id = None, analyzer = None):
		self._id = _id
		self.fields = {}
		self.stored = []
		self.summary = None
		self.analyzer = analyzer
	
	def field (self, name, value, ftype = "Text", lang = "en", encoding = None, option = [], format = None):
		if self._id and name == "_id":
			raise AssertionError ("_id field should be with document initialization")
			
		if type (option) is str:
			option = option.split ()
			
		if ftype == "Fnum":			
			try: 
				format = [int (each) for each in format.split (".", 1)]
				assert len (format) in (1, 2)
			except: 
				raise TypeError ("Invalid Format")
			
			if len (format) == 2:
				a, b = format
			else:
				a, b = fomat [0], 0
			
			a += (b + 1)
			ftype += ("%d.%d" % (a, b))
				
		if type (value) is str:
			value = unistr.makes (value, encoding)
		
		self.fields [name] = (value, ftype, lang, option)			
	
	def documents (self, contents):
		# set one shot, used by schduled indexer
		self.stored = contents
		
	def document (self, content):		
		# any encoding even allowed python objects
		self.stored.append (content)
		
	def snippet (self, text, lang = "en", option = [], encoding = None):
		if type (option) is str:
			option = option.split ()
		# text SHOUL BE string type, DO CEHCK
		self.summary = (unistr.makes (text, encoding), lang, option)
			
	def as_dict (self):
		fields = {}
		if self._id:
			fields ['_id'] = str (self._id)
			
		for k, v in self.fields.items ():
			fields [k] = {'data': v [0], 'type': v [1], 'lang': v [2], 'option': v [3]}
		
		return {
			'snippet': self.summary,
			'documents': self.stored,
			'fields': fields
		}
	
	def as_json (self):
		return json.dumps (self.as_dict ())
		



from . import similarity
import math

class Classifier (similarity.Classifier):
	def match (self, name):
		return name in ("rocchio",)
	
	def revise (self, result):
		s = {}
		for cla, sim in result:
			try: 
				s [cla] += sim
			except KeyError:
				s [cla] = sim	
		result = list (s.items ())
		result.sort (key = lambda x: x [1], reverse = True)
		return result
		



from . import naivebayes
from delune import _delune
import delune
import threading


class Cluster:
	def __init__ (self, f1, threshold, choiceby, verbose = False):		
		self.name = f1
		self.threshold, self.choiceby = threshold, choiceby
		self.terms = {f1: None}
		self.ordered_terms = []
		self.current_socre = 0.0
		self.verbose = verbose
	
	def __contains__ (self, x):
		return (x in self.terms)
		
	def __len__ (self):
		return len (self.terms)
	
	def __str__ (self):
		return "cluster <%s> %s" % (self.name, " ".join (self.get_terms ()))

	def choice_max (self, mis):
		f2s = {}
		for f1 in self.terms:
			for f2, R in mis [f1].items ():				
				if f2 not in self.terms:
					f2s [f2] = None

		cand = None
		for f2 in f2s:
			t = []
			for f1 in self.terms:
				R = mis [f1][f2]
				t.append (R)			
			summi = sum (t)
			if self.choiceby == delune.AVG:
				score = summi / len (t)
			elif self.choiceby == delune.MIN:
				score = min (t)				
			else:
				raise AssertionError ("Unknown choice method: %s" % self.choiceby)
			if cand is None or score > cand [1]:
				cand = (f2, score)
		
		if cand is None:
			return 1
			
		f2, score = cand		
		
		if score <= self.threshold:			
			if self.verbose: self.ordered_terms.append ("%s>>" % f2)
			return f2
		
		self.terms [f2] = None
		self.current_socre = score
		if self.verbose: self.ordered_terms.append (f2)
	
	def get_terms (self):
		if self.ordered_terms:
			return self.ordered_terms
		else:
			return list (self.terms.keys ())			
	
	
class Classifier (naivebayes.Classifier):
	def __init__ (self, reader, analyzer, featset, use_top_features, logger = None):
		self.threshold = 1.0
		self.choiceby = delune.AVG
		self.scoreby = delune.IG
		self.scorer = None
		self.mis_cache = {}
		self.cached = 0		
		self.lock = threading.RLock ()
		naivebayes.Classifier.__init__ (self, reader, analyzer, featset, use_top_features, logger)		
	
	def match (self, name):
		return name == "termcluster"
		
	def setopt (self, **opts):	
		self.threshold = opts.get ("threshold", self.threshold)
		self.choiceby = opts.get ("choiceby", self.choiceby) 
		self.scoreby = opts.get ("scoreby", self.scoreby)
		self.scorer = getattr (_delune, "%smn" % self.scoreby.upper ())
		naivebayes.Classifier.setopt (self, **opts)
		
	def createCluster (self, term):
		return Cluster (term, self.threshold, self.choiceby, self.verbose)
		
	def createTermClusters (self, mem, terms):
		terms = [term for term in terms if term [0] in self.featset]
		if len (terms) <= 1:
			return [terms]
		
		featset = []
		for term, tf  in terms:
			findex = self.featset [term]
			df = self.reader.getDF (mem, term)
			featset.append ((term, findex, df))		
		featset.sort ()
		
		kd = 0
		mis = {}
		min_mi = None
		for term1, x, m in featset [:-1]:
			kd += 1
			if term1 not in mis: 
				mis [term1] = {}
			
			self.lock.acquire ()
			if term1 not in self.mis_cache:
				self.mis_cache [term1] = {}
			self.lock.release ()	
							
			for term2, y, n in featset [kd:]:
				#print (term1, term2)
				if term2 not in mis:
					mis [term2] = {}
				
				try:					
					self.lock.acquire ()
					R = self.mis_cache [term1][term2]					
				except KeyError:						
					self.lock.release ()
					co = self.getCoWord (mem, x, y)					
					if not co:
						R = -15.0
					else:
						R = self.scorer (self.reader.si.getN (), m, n, co)						
					self.lock.acquire ()
					self.mis_cache [term1][term2] = R
					self.cached += 1
					self.lock.release ()
				else:	
					self.lock.release ()
					
				mis [term1][term2] = R
				mis [term2][term1] = R
				
				if min_mi is None or R < min_mi [-1]:
					min_mi = (term1, term2, R)
		
		if self.verbose: self.logger ("co-word %d cached" % len (self.mis_cache))			
		if not mis: return [terms]

		f1, f2, R = min_mi
		if R > self.threshold:
			# single clusters
			if self.verbose: self.logger ("single cluster contains %s and %s with %2.3f" % min_mi)
			return [terms]
		if self.verbose: self.logger ("initial clusters are %s and %s with %2.3f" % min_mi)		

		clusters = {f1: self.createCluster (f1), f2: self.createCluster (f2)}
		dones = []
		while len (dones) < len (clusters):
			for term, cluster in list (clusters.items ()):
				if term in dones: continue					
				unchoiced = cluster.choice_max (mis)
				if unchoiced:
					dones.append (term)
					if unchoiced != 1 and not [x for x in clusters if unchoiced in x]:						
						clusters [unchoiced] = self.createCluster (unchoiced)
					
		terms = dict (terms)		
		term_clusters = []
		olen = len (clusters)
		clusters = self.diminish (clusters)		
		if self.verbose: self.logger ("clusters diminished from %d to %d" % (olen, len (clusters)))
			
		for cluster in clusters:
			if self.verbose: self.logger ("%s\n" % cluster)
			if len (cluster.terms) == 1: continue
			term_clusters.append ([(term, terms [term]) for term in cluster.terms])
		return term_clusters
	
	def diminish (self, clusters):
		cl = list (clusters.values ())
		cl.sort (key = lambda x: len (x.terms))		
			
		deletables = []	
		for i in range (len (cl) - 1):
			for j in range (i+1, len (cl)):				
				has_unique = {}
				for term in cl [i].terms:									
					if term not in cl [j].terms:
						has_unique [term] = None
				
				if not has_unique or (
					i + 1 == j 
					and len (cl [j]) - len (cl [i]) <  len (cl [j]) * 0.1
					and len (has_unique) <= max (1, len (cl [j]) * 0.1)
				):
					# merge cluster
					for newtrem in has_unique:
						cl [j].terms [newtrem] = None
					deletables.append (i)
					break
		
		new_clusters = []
		for i in range (len (cl)):
			if i not in deletables:
				new_clusters.append (cl [i])
		
		return new_clusters	
		
	def cluster (self, mem, qs, lang = "un", terms = None):
		if not terms:
			terms = self.getFeatures (mem, qs, lang)
		return self.createTermClusters (mem, terms)
				
	def guess (self, *args, **kargs):
		raise AttributeError



import threading
import time
import random
import math

gf = open (r"G:\database\classification_Log.LDF", "rb")
glock = threading.Lock ()

READ = 0

def run ():
	global READ
	f = open (r"G:\database\asean_Log.LDF", "rb")
	for i in range (100):
		time.sleep (random.randrange (2))
		f.seek (random.randrange (200000000, 300000000))		
		for j in range (10):
			k = [math.log (k) for k in range (1, 1000)]
			d = f.read (65536)
			glock.acquire ()
			READ += len (d)
			glock.release ()
	f.close ()

def run2 ():
	global gf, glock, READ
	for i in range (100):
		time.sleep (random.randrange (2))
		sp = random.randrange (200000000, 300000000)
		for j in range (10):
			k = [math.log (k) for k in range (1, 1000)]
			glock.acquire ()
			gf.seek (sp + (j * 65536))			
			d = gf.read (65536)
			READ += len (d)
			glock.release ()
	

s = time.time ()	
for i in range (4):
	f = threading.Thread (target = run)
	f.start ()
	
while threading.activeCount () > 1:
	time.sleep (0.1)

print(time.time () - s, READ)

READ = 0
s = time.time ()	
for i in range (4):
	f = threading.Thread (target = run2)
	f.start ()
	
while threading.activeCount () > 1:
	time.sleep (0.1)

print(time.time () - s, READ)

gf.close ()



from delune import _delune
from ...util import util
import os

class SegmentNotOpened (Exception): pass

class Segment:
	def __init__ (self, home = None, seg = -1, mode = 'r', plock = None, version = 1):		
		self.numDoc = 0
		
		self.opened = 0
		self.mode = mode
		self.plock = plock
		self.version = version
		
		self.mutex_id = -1
		
		self.fn_del_modtime = -1
		if seg > -1:
			self.open (home, seg)
		else:
			self.home = home
			self.seg = seg
	
	def has_deleted (self):
		return self.rd is not None
		
	def get_open_flag (self):	
		if self.mode in ("r", "m"):
			fopen_flag = os.O_RDONLY
		else:
			fopen_flag = os.O_WRONLY | os.O_CREAT	
		
		if os.name == "nt":
			fopen_flag |= os.O_BINARY
			
		return fopen_flag
			
	def open (self, home, seg):
		self.home = home
		self.seg = seg
		
		bfn = os.path.join (self.home, str (seg) + ".")		
		fopen_flag = self.get_open_flag ()
					
		fn_tis = bfn + "tis"
		fn_tii = bfn + "tii"
		fn_fdi = bfn + "fdi"
		fn_fda = bfn + "fda"
		fn_tfq = bfn + "tfq"
		fn_prx = bfn + "prx"
		fn_smp = bfn + "smp"
		
		for path in (fn_tis, fn_tii, fn_fdi, fn_fda, fn_tfq, fn_smp):
			self.check_file (path)
		
		self.fs_tis = os.open (fn_tis, fopen_flag)
		self.fs_tii = os.open (fn_tii, fopen_flag)
		self.fs_fdi = os.open (fn_fdi, fopen_flag)
		self.fs_fda = os.open (fn_fda, fopen_flag)
		self.fs_tfq = os.open (fn_tfq, fopen_flag)
		self.fs_prx = os.open (fn_prx, fopen_flag)
		self.fs_smp = os.open (fn_smp, fopen_flag)
		
		bmode = self.mode.encode ("utf8")
		self.ti = _delune.TermInfo (self.fs_tii, self.fs_tis, bmode, self.version)
		self.ti.initialize ()
		
		self.tf = _delune.Posting (self.fs_tfq, self.fs_prx, bmode, self.version)
		self.tf.initialize ()
		
		self.fd = _delune.Document (self.fs_fdi, self.fs_fda, bmode, self.version)		
		self.fd.initialize ()
		
		self.sm = _delune.SortMap (self.fs_smp, bmode, self.version)
		self.sm.initialize ()
		
		self.rd = None
		self.load_deleted ()
				
		self.numDoc = self.si.getSegmentNumDoc (self.seg)
		self.opened = 1		
	
	def get_delfile_modtime (self):
		if self.plock: self.plock.acquire ()		
		try:
			mtime = os.stat (os.path.join (self.home, "%s.del" % self.seg)).st_mtime
		except (IOError, OSError):
			mtime = -1
		if self.plock: self.plock.release ()
		return mtime
	
	def reload_deleted (self):
		# in plock acquired
		new_modtime = self.get_delfile_modtime ()
		if new_modtime == self.fn_del_modtime:				
			return False
			
		if self.rd:
			self.rd.close ()
			self.rd = None
		
		if new_modtime == -1: # .del was deleted
			return True
			
		self.load_deleted ()
		return True
						
	def load_deleted (self):
		fn_del = os.path.join (self.home, "%s.del" % self.seg)
		if not os.path.isfile (fn_del): return			
		if (os.stat (fn_del).st_size > 0):
			if not self.rd:
				self.rd = _delune.BitVector (self.version)
			
			if self.plock: self.plock.acquire ()			
			try:
				self.rd.fromFile (fn_del)
				#f = open (fn_del, "rb")
				#self.rd.fromFile (f)
				#f.close ()				
				self.fn_del_modtime = self.get_delfile_modtime ()
			finally:
				if self.plock: self.plock.release ()	
	
	def commit_deleted (self):
		if self.plock: self.plock.acquire ()	
		try:
			self.create_bitvector ()
			self.rd.toFile (os.path.join (self.home, "%s.del" % self.seg))
			#f = open (, "wb")
			#self.rd.toFile (f)
			#f.close ()
			self.fn_del_modtime = self.get_delfile_modtime ()
		finally:
			if self.plock: self.plock.release ()
	
	def check_segment (self):
		pass
		
	def create_bitvector (self):
		if self.rd: return
		self.rd = _delune.BitVector (self.si.version)
		self.rd.create (self.si.getSegmentNumDoc (self.seg))
	
	def check_file (self, path):
		#if self.mode in ("r", "m"):
		#	assert (os.stat (path).st_size > 0)		
		if self.mode in ("w",):
			if os.path.isfile (path):
				os.remove (path)
		
	def close (self):		
		if not self.opened: 
			return
		
		if self.mode == "w":
			self.fd.commit ()
			self.sm.commit ()
			self.ti.commit ()
			self.tf.commit ()
			os.fsync (self.fs_tis)
			os.fsync (self.fs_tii)
			os.fsync (self.fs_fdi)
			os.fsync (self.fs_fda)
			os.fsync (self.fs_tfq)
			os.fsync (self.fs_prx)
			os.fsync (self.fs_smp)
			
		self.ti.close ()
		self.tf.close ()	
		self.fd.close ()
		self.sm.close ()
		
		os.close (self.fs_tis)
		os.close (self.fs_tii)
		os.close (self.fs_fdi)
		os.close (self.fs_fda)
		os.close (self.fs_tfq)
		os.close (self.fs_prx)
		os.close (self.fs_smp)
		
		if self.rd:
			self.rd.close ()
			self.rd = None
			
		self.opened = 0


class DocumentSegment (Segment):
	def open (self, home, seg):
		self.home = home
		self.seg = seg
		
		bfn = os.path.join (self.home, str (seg) + ".")		
		fopen_flag = self.get_open_flag ()
					
		fn_fdi = bfn + "fdi"
		fn_fda = bfn + "fda"
		
		for path in (fn_fdi, fn_fda):
			self.check_file (path)
			
		self.fs_fdi = os.open (fn_fdi, fopen_flag)
		self.fs_fda = os.open (fn_fda, fopen_flag)
		
		self.fd = _delune.Document (self.fs_fdi, self.fs_fda, self.mode.encode ("utf8"))		
		self.fd.initialize ()
				
		self.opened = 1		
	
	def add_content (self, content, summary = ""):
		self.fd.write ([util.serialize (each) for each in content], summary)
		
	def close (self):
		if not self.opened: 
			return
			
		if self.mode == "w":
			self.fd.commit ()
		self.fd.close ()
		os.close (self.fs_fdi)
		os.close (self.fs_fda)
		
		self.opened = 0



from .. import colsetup
from . import classifier, trainer
from delune.cluster import termCluster

class Segments (colsetup.Segments):
	def __init__ (self, alias = "", version = None):
		colsetup.Segments.__init__ (self, alias, version)
		self.N = 0
		self.pools = []		
		self.numpool = 0
		self.numvoca = 0
		self.features = {}
		self.parameters = {"build": (-1, -1), "train": {}}
		

class Model (colsetup.CollectionSetup):
	exts = ["cfq", "cof", "coi", "fii", "fis", "cfi"]
	segment_class = Segments
	
	def __init__ (self, indexdir, mode = 'r', analyzer = None, logger = None, plock = None, ident = None, version = 1):
		# always MODIFY mode
		colsetup.CollectionSetup.__init__ (
			self, indexdir, mode == 'c' and 'w' or mode, analyzer, logger, plock, ident, version
		)
		if mode == "c":
			self.rebuild = True
		
	def get_learner (self, **karg):
		self.setopt (**karg)
		return trainer.Trainer (self)
	
	def get_classifier (self, **karg):
		self.setopt (**karg)
		return classifier.Classifier (self)
	
	def get_termcluster (self, **karg):
		self.setopt (**karg)
		return termCluster.TermCluster (self)
		



from . import tfidf
import math

class Classifier (tfidf.Classifier):
	def __init__ (self, reader, analyzer, featset, use_top_features, logger = None):
		tfidf.Classifier.__init__ (self, reader, analyzer, featset, use_top_features, logger)
		self.rfeatset = {}
		for k, v in self.featset.items ():
			self.rfeatset [v] = k
		self.fid_type = None	
		
	def match (self, name):
		return name in ("similarity", "sim")
	
	def check_type (self, ostream):
		if type (ostream [0]) is int:
			self.fid_type = True
		else:
			self.fid_type = False
						
	def guess (self, mem, qs, lang = "un", cond = ""):
		terms = self.getFeatures (mem, qs, lang)
		if not terms: return []
		
		K = 2
		A = 0.0
		d = []
		result = []
		comp = {}
		tfc = {}
		TF = 0
		
		for term, tf in terms:
			TF += math.pow (tf, 2.0)
		TF = math.sqrt (TF)

		for term, tf in terms:
			try: 
				tfv = tfc [tf]
			except KeyError:
				tfv = float (tf) / TF
				tfc [tf] = tfv
			df = self.reader.getDF (mem, term)
			fid = self.featset [term]
			comp [fid] = tfv * self.reader.getIDF (df)
			A += math.pow (comp [fid], 2.0)
		A = math.sqrt (A)

		res = self.query (terms, cond, self.fetchcount, lang)
		if res ["total"] == 0:
			return []
		
		if self.fid_type is None:			
			self.check_type (res ["result"][0][0][1])
		
		for row in res ["result"]:
			cla, ostream = row [0][:2]			
			if not ostream: continue
				
			B = C = 0.0
			TF = 0
			tfc = {}
			
			if self.fid_type:
				stream = []			
				for i in range (0, len (ostream), 2):
					if ostream [i] in self.rfeatset:
						stream.append ((ostream [i], ostream [i + 1]))
			else:
				stream = [(self.featset [term], tf) for term, tf in ostream if term in self.featset]
					
			if not stream: continue	
				
			for fid, tf in stream:
				TF += math.pow (tf, 2.0)
			if not TF:continue
			TF = math.sqrt (TF)
			
			for fid, tf in stream:
				try: 
					tfv = tfc [tf]
				except KeyError:
					tfv = float (tf) / TF
					tfc [tf] = tfv
				
				df = self.reader.getDF (mem, self.rfeatset [fid])				
				tfidf = tfv * self.reader.getIDF (df)
				B += math.pow (tfidf, 2.0)
			
				if fid in comp:
					C += comp [fid] * tfidf
			
			if B:
				result.append ((cla, C / (A * math.sqrt (B))))

		return self.revise (result)
	
	def revise (self, result):
		result.sort (key = lambda x: x [1], reverse = True)
		return result



import os
import collections

_ntuple_diskusage = collections.namedtuple('usage', 'total used free')

if hasattr(os, 'statvfs'):  # POSIX
	def disk_usage(path):
		st = os.statvfs(path)
		free = st.f_bavail * st.f_frsize
		total = st.f_blocks * st.f_frsize
		used = (st.f_blocks - st.f_bfree) * st.f_frsize
		return _ntuple_diskusage(total, used, free)

elif os.name == 'nt':	   # Windows		
	import ctypes
	import sys
	
	def disk_usage(path):
		_, total, free = ctypes.c_ulonglong(), ctypes.c_ulonglong(), ctypes.c_ulonglong()
		if sys.version_info >= (3,) or isinstance(path, unicode):
			fun = ctypes.windll.kernel32.GetDiskFreeSpaceExW
		else:
			fun = ctypes.windll.kernel32.GetDiskFreeSpaceExA
		ret = fun(path, ctypes.byref(_), ctypes.byref(total), ctypes.byref(free))
		if ret == 0:
			raise ctypes.WinError()
		used = total.value - free.value
		return _ntuple_diskusage(total.value, used, free.value)

else:
	raise NotImplementedError("platform not supported")

if __name__ == "__main__":    	
	du = disk_usage (r"C:\")
	print ("%d GB Free" % du.free / 1024 / 1024 / 1024)





from ... import _delune
from ...searcher.segment import segment
import os

class Segment (segment.Segment):
	def __init__ (self, home, seg = 0, mode = 'r', plock = None, version = 1):
		self.home = home
		self.ident = "segments"
		self.seg = seg
		self.plock = plock
		self.version = version
		
		self.__opened = 0
		self.__cotermdb = 0
		self.log = None
		self.trace = None
		self.mode = mode
			
	def get_base_path (self):
		return os.path.join (self.home, str (self.seg)  + ".")
		
	def open (self):
		bfn = self.get_base_path ()
		
		if self.mode in ("r", "m"):
			fopen_flag = os.O_RDONLY
		else:
			fopen_flag = os.O_WRONLY | os.O_CREAT
				
		if os.name == "nt":
			fopen_flag |= os.O_BINARY

		self.fs_tis = os.open (bfn + "fis", fopen_flag)
		self.fs_tii = os.open (bfn + "fii", fopen_flag)
		self.fs_tfq = os.open (bfn + "cfq", fopen_flag)
		
		bmode = self.mode.encode ("utf8")
		self.ti = _delune.TermInfo (self.fs_tii, self.fs_tis, bmode, self.version)
		self.ti.initialize ()
		self.tf = _delune.DBInt (self.fs_tfq, bmode, self.version)
		self.tf.initialize ()
		
		try:
			self.fs_cof = os.open (bfn + "cof", fopen_flag)
			self.fs_coi = os.open (bfn + "coi", fopen_flag)
		except OSError:
			pass			
		else:
			self.co = _delune.DBInt (self.fs_cof, bmode, self.version)
			self.co.initialize ()
			self.ci = _delune.DBInt (self.fs_coi, bmode, self.version)
			self.ci.initialize ()		
			self.__cotermdb = 1
		self.__opened = 1
		
	def close (self):
		if not self.__opened:
			raise segment.SegmentNotOpened

		if self.mode == "w":
			self.ti.commit ()
			self.tf.commit ()
			os.fsync (self.fs_tis)
			os.fsync (self.fs_tii) 
			os.fsync (self.fs_tfq)
		
		self.ti.close ()
		self.tf.close ()
		os.close (self.fs_tis)
		os.close (self.fs_tii)
		os.close (self.fs_tfq)
		
		if self.__cotermdb:
			self.co.close ()
			self.ci.close ()
			if self.mode == "w":
				os.fsync (self.fs_cof)
				os.fsync (self.fs_coi)			
			os.close (self.fs_cof)
			os.close (self.fs_coi)
		
		self.__opend = 0



import operator, math
from functools import reduce

#---------------------------------------------------------------------------
# feature selector factory
#---------------------------------------------------------------------------
class FeatureSelectror (object):
	def __init__ (self, tokenCount, docCount, total_tokenCount, total_docCount, total_poolCount):
		self.docCount = docCount
		self.tokenCount = tokenCount
		self.total_docCount = total_docCount
		self.total_poolCount = total_poolCount		
		self.total_tokenCount = total_tokenCount		
	
	def __repr__ (self):
		return '<Base Selector: %d, %d>' % (self.total_docCount, self.total_poolCount)
	
	def __call__ (self, freq, appr):
		return self.compute (freq, appr)
		
	def compute (self, freq, appr):
		I = []		
		sum = reduce (operator.add, appr)		
		for i in range (self.total_poolCount):
			A = appr [i]
			B = sum - A
			C = self.docCount [i] - A
			D = self.total_docCount - (A + B + C)
			result = self.claculate (A, B, C, D)
			I.append (result)
		return self.method (I)
	
	def method (self, I):
		return self.max (I)	
		
	def max (self, I):
		return max (I)
		
	def sd (self, I):
		avg = self.avg (I)
		return math.sqrt (reduce (operator.add, [(x - avg) ** 2 for x in I]) / self.total_poolCount)
	
	def cv (self, I):
		avg = self.avg (I)
		sd = self.sd (I)
		return sd / avg
		
	def avg (self, I):
		return (reduce (operator.add, I)) / len (I)
	
	def asis (self, I):
		return I	
		
class mutual_information (FeatureSelectror):
	def method (self, I):
		return self.sd (I)	
		
	def claculate (self, A, B, C, D):		
		"""
		      1 + (A x N)
		log ---------------
		     (A+C) x (A+B)	
		"""
		r= math.log (1 + float (A * self.total_docCount) / ((A + C) * (A + B)))		
		return r

class chi2statistics (FeatureSelectror):
	def method (self, I):
		return self.max (I)
		
	def claculate (self, A, B, C, D):		
		"""
		      N x (AD - CB)^2
		-----------------------------
		(A+C) x (B+D) x (A+B) x (C+D)	
		"""
		upper = self.total_docCount * ((A * D - C * B) ** 2)
		lower = (A+C) * (B+D) * (A+B) * (C+D)		
		return float (upper) / lower

class dice_coefficient (FeatureSelectror):
	def claculate (self, A, B, C, D):
		return (2. * A) / (2 * A + B + C)		

class jaccard_coefficient (FeatureSelectror):		
	def claculate (self, A, B, C, D):
		return float(A) / (A + B + C)
	
class category_frequency (FeatureSelectror):
	def compute (self, freq, appr):		
		CF = len ([_f for _f in freq if _f])
		result = []
		for i in range (self.total_poolCount):
			if not freq [i]:
				result.append (None)
			else:
				W = freq[i] * (math.log (self.total_poolCount) - math.log (CF) + 1)
				result.append (W)
		return result
	
class document_frequency (category_frequency):
	def compute (self, freq, appr):
		DF = reduce (operator.add, appr)
		result = []
		for i in range (self.total_poolCount):
			if not freq [i]:
				result.append (None)
			else:
				W = freq[i] * (math.log (self.total_docCount) - math.log (DF) + 1)
				result.append (W)
		return result

class oddsratio (category_frequency):
	def compute (self, freq, appr):
		"""
		      P(W|C1) (1-P(W|C2))
		log -----------------------
		      (1-P(W|C1)) P(W|C2)
		      
		odds(W) = (1/n^2) / (1 - 1/n^2) ; p(W) = 0
		        = (1 - 1/n^2) / (1/n^2); P(W) = 1
		        = P(W) / (1 - P(W)) ; P(W) != 0 and P(W) != 1 
		"""
		result = []
		T = reduce (operator.add, freq)
		for i in range (self.total_poolCount):
			targetClassSize = self.tokenCount [i]
			otherClassSize = (self.total_tokenCount - self.tokenCount [i])
			count1 = float (freq [i])
			count2 = float (T - freq [i])
			P_pos = float(count1) / targetClassSize
			P_neg = float(count2) / otherClassSize
			odds_pos = self.calculate (P_pos, count1, targetClassSize)
			odds_neg = self.calculate (P_neg, count2, otherClassSize)
			odds_ratio = math.log (odds_pos / odds_neg)
			result.append (odds_ratio)
		return result
	
	def calculate (self, P, count, size):
		if P == 0.: 
			odds = (1. / size ** 2) / (1 - 1. / size ** 2)
		elif P == 1.: 
			odds = (1 - 1. / size ** 2) / (1. / size ** 2)
		else: 
			odds = (float (count) / size) / (1 - (float (count) / size))
		return odds

class freqlogp (oddsratio):
	def compute (self, freq, appr):
		"""
		                P(W|C1)
		Freq(W) x log -----------
		                P(W|C2)		
		"""
		result = []
		T = reduce (operator.add, freq)
		for i in range (self.total_poolCount):
			targetClassSize = self.tokenCount [i]
			otherClassSize = (self.total_tokenCount - self.tokenCount [i])
			count1 = float (freq [i])
			count2 = float (T - freq [i])
			P_pos = float(count1) / targetClassSize
			P_neg = float(count2) / otherClassSize
			P_pos = self.calculate (P_pos, count1, targetClassSize)
			P_neg = self.calculate (P_neg, count2, otherClassSize)
			freq_log_p = count1 * math.log (P_pos / P_neg)
			result.append (freq_log_p)
		return result
	
	def calculate (self, P, count, size):
		if P == 0.: 
			prob = (1. / size ** 2)
		elif P == 1.: 
			prob = (1 - 1. / size ** 2)
		else: 
			prob = (float (count) / size)
		return prob
		
class information_gain (FeatureSelectror):
	def compute (self, freq, appr):
		"""
		G(t) = -SIGMA P(Ci) * log P(Ci)
			   +P(t) * SIGMA P(Ci|t) * log P(Ci|t)
			   +P(-t) * SIGMA P(Ci|-t) * log P(Ci|-t)	
		"""
		TOTAL_APPR = reduce (operator.add, appr)
		TOTAL_FREQ = reduce (operator.add, freq)
		pc = 0.
		pt_pos = 0.
		pt_neg = 0.		
		for i in range (self.total_poolCount):
			pc += self.docCount [i] / self.total_docCount
			pt_pos += appr [i] / TOTAL_APPR
			pt_neg += (self.docCount [i] - appr[i]) / (self.total_docCount - TOTAL_APPR)
		
		IG = -1. * self.calculate (pc) \
		+ (TOTAL_FREQ / self.total_tokenCount) * self.calculate (pt_pos) \
		+ ((self.total_tokenCount - TOTAL_FREQ) / self.total_tokenCount) * self.calculate (pt_neg)
		return IG
		
	def calculate (self, prob):
		if not prob: return 0.
		return prob * math.log (prob)	
		



import sys
import os
import re
import glob
import zlib
import types
try: 
	import xmlrpc.client
	xmlrpcclient = xmlrpc.client.Server
except ImportError:
	import xmlrpclib
	xmlrpcclient = xmlrpclib.Server
	
from datetime import datetime
import time
import delune
from .. import _delune
from .segment import indexwriter
from .segment import segmentwriter
from .segment import segmentreader
from .segment import segmentmerger
from delune.lib import flock

class ExitNow (Exception):
	pass

class Indexer:
	def __init__ (self, si):
		self.si = si
		self.setMergeFactor (self.si.getopt ("max_segments", 10))
		self._force_merge = self.si.getopt ("force_merge", False)
		self._max_memory = self.si.getopt ("max_memory", 10000000)	# 10M	
		self._optimize = self.si.getopt ("optimize", True)	# 10M	
		self._indexed = 0
		self._initialized = False
		self.writer = None
		self.autocommit = True
		os.umask (0o113)
	
	def set_autocommit (self, flag):
		self.autocommit = flag
		
	def setMergeFactor (self, factor):
		self._merge_factor = factor
		self._merge_factor_onindex = factor * 3
			
	def init (self, newsegment = 1):
		self.si.log ("initializing indexer...", "info")
		if not self.isIndexable ():
			raise ExitNow("`%s' not Indexable" % self.si.getAlias ())
		#self.isDuplicating ()
		self.maybeMerge (self._merge_factor_onindex)
		self.si.lock.lock ("index", str (os.getpid ()))
		if newsegment:
			self.writer = indexwriter.IndexWriter (self.si, self.si.analyzer)
		self._initialized = True
	
	def maybeAbort (self):
		if self.si.lock.islocked ("abort"):
			self.si.lock.unlock ("abort")
			raise ExitNow("`%s' was aborted indexing" % self.si.getAlias ())
		
	def isDuplicating (self):
		while self.si.lock.isplocked ("duplicate"):
			self.si.log ("index is under duplicating, wait 10 sec.", "info")
			time.sleep (10)			
			self.maybeAbort ()

	def isIndexable (self, force = 0):
		if self.si.lock.islocked ("index"):
			if not force and flock.isCurrentProcess (int (self.si.lock.lockread ("index")), 'python'):
				self.si.log ("already under indexing, terminated", "info")
				return 0
		
		if self.si.lock.isplocked ("duplicate"):
			return 0
			
		self.si.lock.unlock ("index")
		self.si.lock.unlock ("merge")
		self.si.lock.unlock ("panic")
		return 1
	
	def truncate (self):
		all_segments = list(self.si.getSegmentInfo ().keys ())
		self.si.log ("truncate all segments...", "info")
		self.mergeSegment (all_segments, truncate = True)
		
	def merge (self):
		all_segments = list(self.si.getSegmentInfo ().keys ())		
		if len (all_segments) == 1 and not self.si.fs.hasDeleted (all_segments [0]):
			return
			
		self.si.log ("merging all segments...", "info")	
		filesizeinfo = {}
		for seg in all_segments:
			if self.merge_and_check (filesizeinfo, seg):
				self.si.log ("file size reaches limit, merging failed", "fail")
				return		
		self.mergeSegment (all_segments)

	def mergeSegment (self, segs, truncate = False):
		t = self.si.getSegmentInfo ()
		t = list(t.items ())
		t.sort (key = lambda x: x[0], reverse = True)

		self.si.log ("merging segments %s" % str (segs))
		self.si.lock.lock ("merge", str (os.getpid ()))
		merger = segmentmerger.SegmentMerger (self.si, truncate)

		try:
			for seg in segs:
				merger.addSegment (seg)
			merger.merge ()

		finally:
			merger.close ()
			self.si.lock.unlock ("merge")
	
	def merge_and_check (self, base, target_segment):
		for k, v in list(self.si.fs.getsegmentsize (target_segment).items ()):
			try: base [k] += v
			except KeyError: base [k] = v
		
		maxsize = max (base.values ())
		if maxsize > delune.LIMIT_SEGMENTSIZE:
			self.si.log ("found LIMIT FILESIZE (%d Bytes), breaking" % maxsize)
			return True
		return False
		
	def maybeMerge (self, merge_factor, force = 0):
		SegmentInfo = self.si.getSegmentInfo ()
		if len (SegmentInfo) <= 2: return
					
		SegmentInfoItems = list(SegmentInfo.items ())
		SegmentInfoItems.sort (key = lambda x: x[0], reverse = True)
		
		seg0 = SegmentInfoItems [0][0]
		seg1 = SegmentInfoItems [1][0]
		if max (self.si.fs.getsegmentsize (seg0).values ()) > delune.LIMIT_FILESIZE: return
		if max (self.si.fs.getsegmentsize (seg1).values ()) > delune.LIMIT_FILESIZE: return
		
		if self.si.lock.islocked ("write"): # deleting documents
			self.si.log ("index writing by another process, wait for 60 sec.")
			for i in range (60): time.sleep (1)
		
		filesizeinfo = self.si.fs.getsegmentsize (seg0)
		self.merge_and_check (filesizeinfo, seg1)
		targetSegments = [seg0, seg1]
		segmentIndex = 2
		isOver = False
		while segmentIndex < len (SegmentInfoItems):
			seg, numDoc = SegmentInfoItems [segmentIndex]
			if max (self.si.fs.getsegmentsize (seg).values ()) > delune.LIMIT_FILESIZE: 
				self.si.log ("found BIG FILESIZE (%d Bytes), breaking" % max (self.si.fs.getsegmentsize (seg).values ()))
				break
			targetSegments.append (seg)
			if self.merge_and_check (filesizeinfo, seg):
				isOver = True
				break			
			segmentIndex += 1
			
		if isOver:
			self.si.log ("start merging segments because reaching size limit...")
			return self.mergeSegment (targetSegments)
			
		if not force and len (SegmentInfo) <= merge_factor:
			return
		
		filesizeinfo = self.si.fs.getsegmentsize (seg0)
		self.merge_and_check (filesizeinfo, seg1)
		targetSegments = [seg0, seg1]
		segmentIndex = 2
		mergeDocCount = SegmentInfoItems [0][1] + SegmentInfoItems [1][1]
		while segmentIndex < len (SegmentInfoItems):
			seg, numDoc = SegmentInfoItems [segmentIndex]
			if max (self.si.fs.getsegmentsize (seg).values ()) > delune.LIMIT_FILESIZE: 
				self.si.log ("found BIG FILESIZE (%d Bytes), breaking" % max (self.si.fs.getsegmentsize (seg).values ()))
				break
			if self.merge_and_check (filesizeinfo, seg):
				targetSegments.append (seg)
				self.si.log ("start merging segments because reaching size limit while optimizing...")
				return self.mergeSegment (targetSegments)			
			if numDoc > mergeDocCount <= merge_factor: break
			targetSegments.append (seg)
			mergeDocCount += numDoc			
			segmentIndex += 1
			
		self.si.log ("start merging segments because optimizing...")
		self.mergeSegment (targetSegments)
			
	def optimize (self, force = 0):
		if not self._optimize: 
			return
		
		if force or self._force_merge:
			self.si.log ("optimizing segments forcely...")
			self.merge ()

		else:
			self.si.log ("optimizing segments...")
			self.maybeMerge (self._merge_factor)			

	def commit (self, final = 0):
		self.si.log ("commit segment...")
		if self.writer:
			self.writer.flush (final)
		# replace new writer
		if not final:
			self.maybeMerge (self._merge_factor_onindex)

	def close (self, err = 0, optimize = 0):
		self.si.log ("closing indexer...")
		self.commit (1)
		self.optimize (optimize)
		
		if self.writer:
			self.writer.close ()
			
		self.si.lock.unlock ("index")
		self.si.log ("indexer closed")
		self.si.close ()
		self.si = None
	
	def add_document (self, document):
		if not self._initialized: 			
			self.init ()
		self.maybeAbort ()
		
		wdoc = self.writer.newWDocument ()
		
		wdoc.setData (document.stored)
		if document.summary:
			wdoc.addSummary (*document.summary)
		
		# exception handling per fields
		for name, (val, ftype, lang, option) in list(document.fields.items ()):			
			try:
				wdoc.addField (ftype, name, val, lang, option)
			except:
				self.si.trace ()
				self.si.log ("%s %s indexing failed %s" % (ftype, name, val), "fail")
		
		self.writer.addWDocument (wdoc)
		
		self._indexed += 1		
		if self._indexed % 100 == 0:
			self.si.log ("indexed %d documents, %d terms" % (self._indexed, self.writer.segment.termTable.numterm))
		if self._indexed % 1000 == 0:
			self.si.log ("estimated %3.2f MB data allocated" % (self.writer.getMemUsage () / 1048576.0,))
		if self.autocommit and self.is_memory_over ():
			self.commit ()
	
	def is_memory_over (self):
		if not self.writer:
			return False
		return self.writer.getMemUsage () >= self._max_memory
		
	def delete_documents_by_query (self, qs):
		if not qs: return		
		s = xmlrpcclient.Server ("http://" + self.si.index_server)
		s.delune.util.delete (self.si.getAlias (), qs)
		s.delune.util.commit (self.si.getAlias ())




import re
from . import typeinfo
import delune


def translate (analyzer, text, lang, option):
	if "strip_html" in option:
		temp = text
		text = analyzer.formalize (analyzer.strip_html (temp))
	if "formalize" in option:
		text = analyzer.formalize (text)
	if "lowercase" in option:
		text = text.lower ()
	if "stem" in option:
		text = analyzer.term (text, lang)
		if text is None:
			text = []

	return text	
		
class FieldWriter:
	def __init__ (self, segment, analyzer = None):
		self.segment = segment
		self.analyzer = analyzer
		self.tcache = {}

	#---------------------------------------------------------------------------
	# add posting
	#---------------------------------------------------------------------------
	def indexText (self, type, docid, num, name, value, lang, option):
		if self.analyzer:
			value = self.analyzer.analyze (translate (self.analyzer, value, lang, option), lang, type)
		
		if not value:
			self.segment.writeNorm (num, 0)
			return
		
		numTerm = 0
		if not value:
			self.segment.writeNorm (num, 0)
			return
		
		if type == delune.TEXT:
			for term in value:				
				self.segment.writePosting (docid, num, term, len (value [term]), value [term])
				numTerm += len (value [term])

		elif type == delune.TERM:
			for term, tf in value:
				self.segment.writePosting (docid, num, term, tf)
				numTerm += tf

		self.segment.writeNorm (num, numTerm)
	
	def indexFnum (self, type, docid, num, name, value, lang, option):
		if not value: return
		if isinstance (value, str):	
			value = [each.strip () for each in value.split (",")]		
		if not isinstance(value, (list, tuple)):	
			value = [value]
		value = [typeinfo.zfill (type, each) for each in value]
		self.indexField (type, docid, num, name, value, lang, option)

	def indexField (self, type, docid, num, name, value, lang, option):
		if not value: return
		if not isinstance(value, (list, tuple)):
			value = [value]
		for each in value:
			self.segment.writePosting (docid, num, each, 1)

	def indexList (self, ftype, docid, num, name, value, lang, option):
		if not value: return
		if type (value)	is type (""):			
			delim = ","
			if "delim" in option:
				ii = option.index ("delim")
				try:
					delim = option [ii + 1]
				except IndexError:
					pass
			value = value.split (delim)
		
		d = dict ([(_f, None) for _f in [str (x).strip () for x in value] if _f])
		self.indexField (ftype, docid, num, name, list (d.keys ()), lang, option)
	
	def indexStringSort (self, type, docid, num, name, value, option):
		self.segment.writeStringSortKey (num, value, typeinfo.typemap.getsize (type))
			
	def indexInt (self, type, docid, num, name, value, lang, option):
		# itn, bit, coord must be indexed even value is totally non-sense
		try: val = int (value)
		except: val = 0
		self.segment.writeSortKey (num, val, typeinfo.typemap.getsize (type))

	def indexBit (self, type, docid, num, name, value, lang, option):
		if not value:
			val = 0		
		elif isinstance (value, int):
			val = value		
		else:
			try: 
				val = int (value, 2)
			except: 
				try:
					val = int (value)
				except ValueError:
					val = 0					
		self.segment.writeSortKey (num, val, typeinfo.typemap.getsize (type))

	def indexCoord (self, type, docid, num, name, value, lang, option):
		tsize = typeinfo.typemap.getsize (type)
		precision = 10 ** (tsize-2)
		try:
			val = [int ((round (float (x), tsize) + 180.0) * precision) for x in value.split ("/", 1)]
			assert (len (val) == 2)
			if [x for x in val if x > (360 * precision)]:
				raise ValueError
		except:
			val = [0, 0]			
		self.segment.writeSortKey (num, val, tsize)
	
	def __call__ (self, type, docid, num, name, value, lang, option):
		method = getattr (self, "index" + typeinfo.typemap.getmethod (type))
		method (type, docid, num, name, value, lang, option)



import sys
import os

class index_sync_handler:
	def __init__(self, wasc):
		self.wasc = wasc
		
	def match (self, request):
		if request.command.lower() == 'replicate':
			return 1
		return 0
	
	def handle_request (self, request):
		authinfo = request.get_header ('Authorization')
		#if not authinfo or not self.wasc.authorizer (authinfo):
		#	return request.error (403)		
		try:
			request.collector = collector (self, request)
		except:
			self.wasc.se.logger.trace ()	
				
	def continue_request (self, request, filename = ""):
		if not filename:
			return request.error (500)
		
		request.push ("Created %s\n" % filename)
		request.done ()
		

class collector:
	def __init__ (self, handler, request):
		self.handler = handler
		self.request = request
		
		cl = request.get_header ('content-length')
		if not cl:
			request.error (411)
						
		else:
			cl = int(cl)
			if not cl: 	self.found_terminator()
			self.request.channel.set_terminator (cl)
			
		self.fn = request.get_header ("se-filename")		
		if not self.fn: 
			request.error (500)
			return
		
		self.ft = request.get_header ("se-filetype")
		if not self.ft: 
			request.error (500)
			return
		
		self.col = request.get_header ("se-index")	
		if not self.col: 
			request.error (500)
			return
		
		if self.ft == "config":			
			self.__create_cfile ()
		elif self.ft == "segmentinfo":
			self.__create_sfile ()
		else:
			self.__create_ifile ()
	
	def collect_incoming_data (self, data):
		self.fp.write (data)

	def found_terminator (self):		
		self.fp.close ()
		self.request.channel.set_terminator ('\r\n\r\n')
		res = self.__commit_file ()
		
		if not res: self.filepath = ""
		self.handler.continue_request (self.request, self.filepath)

	def __get_searcher (self):
		searcher = self.handler.server.se.getsearcher (self.col)
		if not searcher:
			searcher = self.handler.server.se.create_searcher (os.path.join (os.environ ["INSTANCE_HOME"], "etc", "col", self.col))
		return searcher
	
	def __commit_file (self):
		if self.ft == "index":
			return 1
				
		if self.ft == "config":
			nfile = os.path.join (os.environ ["INSTANCE_HOME"], "etc", "col", self.col)
			
		else:
			nfile = self.filepath [:-4]
		
		try:
			try: os.remove (nfile)
			except OSError as why:
				if why [0] != 2:
					raise
					
			os.rename (self.filepath, nfile)
				
		except:
			self.handler.server.se.logger.trace ()
			return 0
			
		return 1
			
	def __create_cfile (self):
		self.filepath = os.path.join (os.environ ["INSTANCE_HOME"], "etc", "col", "#" + self.col + ".tmp")
		self.fp = open (self.filepath, "w")
	
	def __create_sfile (self, col, fn):		
		searcher = self.__get_searcher ()						
		path = searcher.actor.si.fs.getmaster ()
		self.filepath = os.path.join (path, fn + ".tmp")
		self.fp = open (self.filepath, "wb")		
		
	def __create_ifile (self):
		searcher = self.__get_searcher ()		
		segment = self.fn [:-4]
		path = searcher.actor.si.fs.get (segment)
		
		if not path:
			path = searcher.actor.si.fs.new ()
		
		self.filepath = os.path.join (path, fn)
		self.fp = open (self.filepath, "wb")
	
	